# Re-export from discord_slash
from discord_slash.context import SlashContext
from discord_slash.model import SlashCommandOptionType
from discord_slash.utils import manage_commands

# Export own functions and classes
from .client import SlashCommand
from .converters import convert
from .decorators import slash, slash_cog

__version__ = "1.1.1"



from typing import Any, Callable, Dict, Union
import discord

from discord.ext import commands
from discord.ext.commands.errors import BadArgument
from discord_slash import SlashContext


class BadSlashArgument(commands.BadArgument):
    """
    Exception class for bad slash arguments. Raised by `handle_arg` and
    consequently `convert`.

    .. note::
        You should not raise this manually. This should be distinct to the
        converters.
    """
    pass


async def handle_arg(
    ctx: SlashContext,
    key: str,
    value: Any,
    type_: Union[Callable, commands.Converter]
) -> Any:
    """
    Handle an argument and deal with typing.Optional modifiers

    Parameters
    ----------
    ctx : SlashContext
        The context of the argument
    key : str
        The argument name
    value : any
        The value of the argument
    type_ : callable or commands.Converter
        The type, instance of type converter or type converter class

    Returns
    -------
    any
        The handled argument

    Raises
    ------
    BadSlashArgument
        Invalid argument, this occurs when no converter is able to provide a
        non-None response or not raise an error while converting. If a
        typing.Optional is provided, or any object with .__args__ containing
        NoneType, then this method will never raise and will instead return
        None.
    """
    # Is a typing.Optional, typing.Union, etc class
    if hasattr(type_, "__args__"):
        optional = type(None) in type_.__args__
        for item in type_.__args__:  # Iterate through possible types
            if item is None:
                # Don't try to convert with None
                continue

            try:
                # Attempt to convert, this also allows nesting of
                # typing.Optional etc
                new_value = await handle_arg(ctx, key, value, item)

                # Return by default if it's good (should go left to right)
                return new_value
            except Exception as exc:
                # This is optional, so we can skip past it
                pass
        
        if not optional:
            raise BadSlashArgument(message=f"Argument {key} is not of any valid type")
        
        return None
    else:
        if hasattr(type_, "convert"):
            # Check item is instantiated
            if isinstance(type_, type):
                # Instantaniate the class first
                type_converter = type_().convert
            else:
                # Grab the function of the init'd converter
                type_converter = type_.convert
            try:
                return await type_converter(ctx, value)
            except Exception as exc:
                raise BadSlashArgument(f"Failed to convert argument {key}") from exc
        else:
            # Probably not a converter
            return type_(value)
    

def convert(send_on_raise: bool = False, **converters: Dict[str, commands.Converter]):
    """
    Wraps slash commands to perform extra conversions on functions where
    necessary. Uses the same interface as the default command
    converter.

    .. note::
        This should only be used if you really really want to use the old
        slash interface. Otherwise, it's better to just use the decorators
        provided.
    
    Parameters
    ----------
    send_on_raise : bool
        Whether to send an epheremal message with the error message on failure,
        by default False
    **converters : Dict[str, commands.Converter]
        Converters where the name = Converter
    
    """
    def decorator(function):
        async def wrapper(self_or_ctx, *args, **kwargs):
            ctx = self_or_ctx if isinstance(self_or_ctx, SlashContext) else args[0]
            for key, type_ in converters.items():
                # If they're not a string, chances are they've already been
                # converted (member, int, etc)
                if type(kwargs.get(key)) != str:
                    continue
                
                try:
                    kwargs[key] = await handle_arg(ctx, key, kwargs.get(key), type_)
                except BadArgument as exc:
                    if send_on_raise:
                        await ctx.send(str(exc), hidden=True)
                    raise exc

            await function(self_or_ctx, *args, **kwargs)
        wrapper.__annotations__ = function.__annotations__
        return wrapper
    return decorator



from typing import List
from discord_slash import SlashCommand as SlashCommandOriginal

from .decorators import slash


class SlashCommand(SlashCommandOriginal):
    def slash(self, name: str = None, description: str = None, guild_ids: List[int] = None, remove_underscore_keywords: bool = True):
        """
        Add a command to a bot at the top level

        .. note::
            Unlike the original implementation of `SlashCommand.slash`, this
            doesn't require parameter definitions and also allows converters

        Parameters
        ----------
        slash_class : SlashCommand
            The slash object to add this command to
        name : str, optional
            The name of the command, by default None
        description : str, optional
            The description, by default None
        guild_ids : List[int], optional
            The guild IDs to add the command to, by default None
        remove_underscore_keywords : bool, optional
            Whether to remove _ from the end of arguments that would be keywords,
            by default True
        """
        return slash(
            slash_class=super(),
            name=name,
            description=description,
            guild_ids=guild_ids,
            remove_underscore_keywords=remove_underscore_keywords
        )




from typing import List

from discord_slash import SlashCommand, cog_ext

from .converters import convert
from .utils import *


def slash_cog(name: str = None, description: str = None, guild_ids: List[int] = None, remove_underscore_keywords: bool = True):
    """
    Add a command to a cog

    Parameters
    ----------
    name : str, optional
        The name of the command, by default None
    description : str, optional
        The description, by default None
    guild_ids : List[int], optional
        The guild IDs to add the command to, by default None
    remove_underscore_keywords : bool, optional
        Whether to remove _ from the end of arguments that would be keywords,
        by default True
    """
    def decorator(function):
        # Use annotations
        params, converter_params = get_slash_kwargs(
            name, description, guild_ids, remove_underscore_keywords, function)
        return cog_ext.cog_slash(**params)(convert(**converter_params)(function))

    return decorator


def slash(slash_class: SlashCommand, name: str = None, description: str = None, guild_ids: List[int] = None, remove_underscore_keywords: bool = True):
    """
    Add a command to a bot at the top level

    Parameters
    ----------
    slash_class : SlashCommand
        The slash object to add this command to
    name : str, optional
        The name of the command, by default None
    description : str, optional
        The description, by default None
    guild_ids : List[int], optional
        The guild IDs to add the command to, by default None
    remove_underscore_keywords : bool, optional
        Whether to remove _ from the end of arguments that would be keywords,
        by default True
    """
    def decorator(function):
        # Use annotations
        params, converter_params = get_slash_kwargs(
            name, description, guild_ids, remove_underscore_keywords, function)
        return slash_class.slash(**params)(convert(**converter_params)(function))

    return decorator



import inspect
import keyword
from typing import Any, List, Literal, Union

from discord.ext import commands
from discord.ext.commands.errors import CommandError
from discord_slash.context import SlashContext
from discord_slash.model import SlashCommandOptionType
from discord_slash.utils.manage_commands import create_choice, create_option


class InvalidParameter(CommandError):
    pass


def is_converter(value: Any) -> bool:
    return (isinstance(value, type) and issubclass(value, commands.Converter)) \
        or isinstance(value, commands.Converter)


def is_optional_of(value: Any, i: Any):
    """
    See if `value` is a subclass of `i` hidden by `typing.Optional` or
    `typing.Union`

    Parameters
    ----------
    value : Any
        The potential class
    i : Any
        The class to check against
    """
    return get_root_type(value) == i


def get_root_type(value: Any):
    """
    Gets the root type of an annotation naively

    Parameters
    ----------
    value : Any
        The annotation to check

    Returns
    -------
    type
        A type
    """
    if hasattr(value, "__args__"):
        return get_root_type(value.__args__[0])

    if isinstance(value, type):
        return value
    else:
        return type(value)


def validate_literal_union(union: Any) -> bool:
    """
    Validate a Union parameter. All arguments must be a literal and
    of the same type. This only includes the first parameter

    Parameters
    ----------
    union : Union
        The Union to check

    Returns
    -------
    bool
        Whether it's valid
    """
    if not hasattr(union, "__args__"):
        return False

    are_all_literals = all(getattr(arg, "__origin__", None)
                           == Literal for arg in union.__args__)

    if are_all_literals:
        are_all_same_type = len(
            set(type(arg.__args__[0]) for arg in union.__args__)) == 1
        return are_all_same_type

    return False


def get_slash_command_type(annotation: Any) -> SlashCommandOptionType:
    """
    Get slash command type from an annotation. This considers optionals and
    converters

    Parameters
    ----------
    annotation : Any
        Annotation to check

    Returns
    -------
    SlashCommandOptionType
        The parameter type

    Raises
    ------
    InvalidParameter
        No parameter type could be found
    """
    # If it's a converter or an optional of a converter, it will always be a string
    if issubclass(get_root_type(annotation), commands.Converter):
        return SlashCommandOptionType.STRING

    type_ = SlashCommandOptionType.from_type(get_root_type(annotation))
    if type_ is None:
        raise InvalidParameter(
            f"Parameter: {str(annotation)} does not match any Discord type")

    return type_


def get_slash_kwargs(name: str, description: str, guild_ids: List[int], remove_underscore_keywords: bool, function):
    """
    Get the kwargs required for cog_ext.cog_slash or SlashCommand.slash

    Parameters
    ----------
    name : str
        The name of the command
    description : str
        The description of the command
    guild_ids : List[int]
        List of guild IDs to add the command to
    remove_underscore_keywords : bool
        Whether to remove _ from the end of arguments that would be keywords
    function : any
        The command

    Returns
    -------
    dict
        The arguments as a kwarg
    dict
        The connectors for the parameter names
    """
    # Use annotations
    signature = inspect.signature(function)

    params = dict(
        name=name or function.__name__,
        description=description or "No description",
        guild_ids=guild_ids,
        options=[]
    )
    param_name_mapping = dict()
    converter_params = dict()

    for param_name, parameter in signature.parameters.items():
        annotation = parameter.annotation
        if annotation == SlashContext or param_name == "self":
            continue

        # If it's a keyword with an underscore, then remove the underscore
        # for Discord's sake
        if remove_underscore_keywords and keyword.iskeyword(param_name[:-1]):
            # def_ -> def, in_ -> in
            param_name_mapping[param_name[:-1]] = param_name
            param_name = param_name[:-1]

        # If it's a tuple of a type and a literal, the second argument is a description
        param_description = "No description"
        if getattr(annotation, "__origin__", None) == tuple and len(annotation.__args__) == 2:
            param_description = annotation.__args__[1].__args__[0]
            annotation = annotation.__args__[0]

        # Default to no choices
        choices = None
        # Unions of literals are considered choices, rather than converters
        if validate_literal_union(annotation):
            choices = []
            for literal in annotation.__args__:
                # Literal[1, "My naparam_nameme"] would give choice 1, of name Name
                choice_value = literal.__args__[0]
                if len(literal.__args__) > 1:
                    choice_name = literal.__args__[1]
                else:
                    choice_name = str(choice_value)
                choices.append(create_choice(choice_value, choice_name))
        else:
            # Just use converter params
            converter_params[param_name] = annotation

        params['options'].append(create_option(
            name=param_name,
            description=param_description,
            option_type=get_slash_command_type(annotation),
            required=parameter.default == inspect.Parameter.empty,
            choices=choices
        ))

    params['connector'] = param_name_mapping
    return params, converter_params



from os import path
from setuptools import setup

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(path.abspath(path.dirname(__file__)), 'requirements.txt'), encoding='utf-8') as f:
    install_requires = f.readlines()

setup(name='dpyslash',
      version='1.1.1',
      description='Improves slash commands for Python',
      author='starsflower',
      url='https://github.com/starsflower/discordpy-slash-commands',
      license="MIT",
      classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
      ],
      packages=['pyslash'],
      package_dir={'pyslash': './pyslash'},
      install_requires=install_requires,

      # Description
      long_description=long_description,
      long_description_content_type='text/markdown'
)
